"""
Functions required by Concaterpillar.

"""

from filetools import *
from seqtools import *
from statstools import *
from treetools import *

import cPickle, getopt, glob, re, sys, shutil, threading
import Queue

def brlensetup():
  """Move to the directory BLtest, if it exists, and copy relevant files.

  If the directory BLtest does not exist, it is created.  If the file 
  catsets.pickle exists (from the congruence test), the largest concatenated
  set is assumed to be the one on which the branch length compatibility test
  should be performed.  The sequences from this set are copied into BLtest,
  along with the tree from their concatenation.  If catsets.pickle does not
  exist, all sequence files are copied over, along with the tree file global.tre
  (if it exists), which is assumed to be the tree from concatenation of all
  alignments.  If it doesn't exist, alignments are concatenated, and PHYML is
  run on the global alignment to produce the global tree.
  
  Return:
    list containing names of sequence files copied to the BLtest directory
  
  """
  
  bltestdirs = glob.glob('BLtest*')
  
  try:
    backup = open('catsets.pickle')
    catsets = cPickle.load(backup)
    backup.close()

    setcount = 0
    for set in catsets:
      bldir = 'BLtest-set%03d' % setcount 
      os.mkdir(bldir)
      shutil.copy('set%03d.tre' % setcount, '%s/global.tre' % bldir)
      
      for gene in set:
        shutil.copy('%s' % gene, '%s/%s' % (bldir, gene))

      setcount += 1
      
  except IOError:
    print '\nCouldn\'t find backup from previous concaterpillar run. ' + \
          'Branch length congruence\ntest will be performed on all sequences!\n'      
    sys.stdout.flush()
    
    # make a "fake" catsets, with all genes in same topologically congruent set
    catsets = [getseqlist()] 
    if 'global.tre' not in os.listdir('.'):
      writephy(catseq(getaligns(catsets[0])), 'global.seq')
      buildmultitrees(('global.seq',))

  return catsets    

def brlentest(catsets, setdir, cutoff, nprocs = 1):
  """Execute the branch-length congruence test.

  Parameters:
    catsets -- list of topologically congruent sets (produced by toptest)
    setdir -- "name" for branch-length test run (e.g. BLtest-set000/)
    cutoff -- uncorrected alpha-level for the test
    nprocs -- number of processes to thread off... should be number of 
              processors to use on a shared memory machine
  
  """
  
  resfile = open('%s.ccp' % setdir[:-1].lower(), 'w')
  gammacats = 4
  lastpval = 0
  q = Queue.Queue()
  lock = threading.Lock()
  threadcounter = Counter()
  
  treefile = open('global.tre')
  tree = treefile.readline().strip()
  treefile.close()
  
  nsets = 0
  for set in catsets:
    nsets += len(set)
    
  # current level is number of concatenations that have alredy happened (single-
  # gene sets minus remaining sets plus 1... note incrementation in loop below)
  level = nsets - len(catsets)

  while len(catsets) > 1:
    level += 1
    dlist = os.listdir('.')

    for i in range(len(catsets)):
      if 'set%03d.lnl' % i not in dlist:
        q.put({'tree': tree, 'seqfile': 'set%03d.seq' % i, \
             'gammacats': gammacats})

      for j in range(i + 1, len(catsets)):
        catsetname = 'set%03d%03d.seq' % (i, j)

        # if concatenated file doesn't exist, create it
        if catsetname not in dlist: 
          writephy(catseq(getaligns(('set%03d.seq' % i, 'set%03d.seq' % j))), 
                   catsetname)
        
        if 'set%03d%03d.lnl' % (i, j) not in dlist:
          q.put({'tree': tree, 'seqfile': catsetname, 'gammacats': gammacats})

    for i in range(nprocs):
      t = PhyloThread(q, lock, target=runpuzzlebrlen, name='thread %d' % i,
                      counter=threadcounter)
      t.start()
      
      # wait until thread really starts before starting next thread
      lock.acquire()
      lock.release()
      
    while int(threadcounter) > 0:
      time.sleep(10)

    # dlnL (lnL(A) + lnL(B) - lnL(AB))
    matrix = []

    # diff. in deg. of freedom
    dfmat = []
    for i in range(len(catsets)):
      matrix.append([])
      dfmat.append([])
      for j in range(i + 1, len(catsets)):
        result = calcBLratio(i, j)
        matrix[-1].append(result[0])
        dfmat[-1].append(result[1])
      
    
    row, col = getmin(matrix)
    setA = row
    setB = col + row + 1
    
    # p-val for chi-square test (note that lnL is multiplied by 2)
    pval = 1 - pchisq(2 * matrix[row][col], dfmat[row][col])      
    
    # Correct alpha based on current level... see argument in toptest
    alpha = correctalpha(cutoff, level)
    
    # Only tell the user we're correcting if we NEED to correct
    if pval < cutoff:
      print 'Applying alpha-level correction: new alpha = %.6f' % alpha
      resfile.write('Applying alpha-level correction: new alpha = %.6f\n' % alpha)
      sys.stdout.flush()
      resfile.flush()
    
    if pval < alpha:
      print 'Failed to concatenate %s and %s (p = %.6f).\n' % \
            (str(catsets[setA]), str(catsets[setB]), pval)
      resfile.write('\nFailed to concatenate %s and %s (p = %.6f).\n\n' % \
                    (str(catsets[setA]), str(catsets[setB]), pval))
      resfile.flush()
      break
      
    else: # p-val of chi-square test >= corrected alpha
      print 'Concatenating %s and %s (p = %.6f).\n' % \
            (str(catsets[setA]), str(catsets[setB]), pval)
      resfile.write('Concatenating %s and %s (p = %.6f).\n' % \
                    (str(catsets[setA]), str(catsets[setB]), pval))
      resfile.flush()
      
      # rename and remove necessary sets
      filedict = {}
      for i in ('.seq', '.lnl', '.prm'):
        filedict['set%03d%03d%s' % (setA, setB, i)] = 'set%03d%s' % (setA, i)
      renamefiles(filedict)

      # necessary only if setB is highest-numbered set
      removefiles(glob.glob('set%03d*' % setB))
      removesets(setA)
      removesets(setB)
      renamesets(setB)
      
      catsets[setA].extend(catsets[setB])
      del catsets[setB]
      
  print 'Finished concatenating sets.  The following files contain ' + \
        'the final sets:'
  resfile.write('Finished concatenating sets.  The following files contain ' + \
                'the final sets:\n')

  for i in range(len(catsets)):
    print '%sset%03d.seq (%d genes): %s' % (setdir, i, len(catsets[i]), 
                                                 str(catsets[i]))
    resfile.write('%sset%03d.seq (%d genes): %s\n' % (setdir, i, 
                                                       len(catsets[i]), 
                                                       str(catsets[i])))
  resfile.close()
  
  print ''
  
def buildmatrix(likes):
  """Build a matrix of log-likelihood ratios from a matrix of log-likelihoods.

  The matrix returned is in upper-triangular form, and contains likelihood
  ratios of the form lnL^A_{T_A} + lnL^B_{T_B} - (lnL^A_{T_AB} + lnL^B_{T_AB})

  Parameters:
    likes -- a matrix of pairs of log-likelihoods [lnL^A_{T_A}, lnL^A_{T_AB}]
             with the diagonal missing (not invalid values, these cells are
             just absent)
             
  Return:
    Log-likelihood ratio matrix.
    
  """
  difmat = []
  nsets = len(likes)

  for i in range(nsets):
    row = []
    for j in range(1, nsets):
      # PHYML may estimate a better tree for alignment A from alignment AB
      # (this constrains differences to be at least 0)
      row.append(max(likes[i][j - 1]) - likes[i][j - 1][1])
    difmat.append(row)
    
  for i in range(len(difmat)):
    for j in range(len(difmat[i]) + 1):
      if i == j:
        difmat[i][i:i] = [0.0]

  # add across the diagonal to get the likelihood ratio including both genes
  symmat = []
  for i in range(nsets - 1):
    row = []
    for j in range(i + 1, nsets):
      row.append(difmat[i][j] + difmat[j][i])
    symmat.append(row)
    
  return symmat

def buildmultitrees(infiles, nprocs = 1):
  """Run PHYML on a list of alignment filenames.
  
  Parameters:
    infiles -- list of file names on which to run PHYML
    nprocs -- number of processes to thread off... should be number of 
              processors to use on a shared memory machine
  
  """

  print 'I\'m about to build trees for %d data sets (using phyml).' % (len(infiles))
  
  q = Queue.Queue()
  for i in infiles:
    cmds = {'model': 'JTT', 'seqfile': i}
    q.put(cmds)
  
  lock = threading.Lock()
  threadcounter = Counter()

  for i in range(nprocs):
    t = PhyloThread(name = 'thread %d' % i, jobqueue = q, 
                    lock = lock, target = buildtrees, counter = threadcounter)
    t.start()
    
    # wait until new thread really starts before starting next thread
    lock.acquire()
    lock.release()
    
  while int(threadcounter) > 0:
    time.sleep(10)
    
def buildtrees(cmds):
  """Run PHYML with the parameters specified.
  
  Parameters:
    cmds -- dictionary specifying the sequence file and model to use with PHYML.
  
  """
  
  phymlpath = globals()['phymlpath']
  infile = cmds['seqfile']
  
  # try to run PHYML up to 5 times... this is only to deal with our sick cluster
  nattempts = 5
  for i in xrange(nattempts):
    runphyml(phymlpath, alignname = infile, model = cmds['model'])
    alphafile = open('%s.alpha' % infile[:-4], 'w')
    try:
      alpha = getalpha('%s_phyml_stat.txt' % infile)
      break
    except ConcaterpillarError:
      sys.stderr.write('Somehow, PHYML was killed. Trying again...\n')
      
  else:
    sys.stderr.write('Exceeded number of attempts to run PHYML (%d). ')
    sys.stderr.write('Killing thread.\n')
    sys.exit(1)
      
  if alpha == None:
    sys.stderr.write('There was a problem running PHYML with file %s.\n' % 
                     infile)
    sys.stderr.write('Please check the file and try again.\n\n')
    sys.exit(1)
    
  alphafile.write('%.6f\n' % alpha)
  alphafile.close()
    
  # clean up
  renamefiles({'%s_phyml_tree.txt' % infile: infile[:-4] + '.tre'}) 
  removefiles(('%s_phyml_stat.txt' % infile, '%s_phyml_lk.txt' % infile))
  
  print 'Finished analysis of %s.' % (infile)
  sys.stdout.flush()

def calcBLratio(seqA, seqB):
  """Calculate the log-likelihood ratio for the branch-length test.
  
  Parameters:
    setA: number of the first set
    setB: number of the second set
    
  Return:
    Tuple containing the likelihood ratio and number of degrees of freedom
    
  """
  
  likes = {}
  nparams = {}
  
  setnames = {'A': 'set%03d' % seqA, 'B': 'set%03d' % seqB, \
              'AB': 'set%03d%03d' % (seqA, seqB)}
  
  
  for setkey in setnames:
    lnlfile = open('%s.lnl' % setnames[setkey])
    likes[setkey] = float(lnlfile.readline().strip())
    lnlfile.close()
      
    paramfile = open('%s.prm' % setnames[setkey])
    nparams[setkey] = int(paramfile.readline().strip())
    paramfile.close()
       
  result = likes['A'] + likes['B'] - likes['AB']
  df = nparams['A'] + nparams['B'] - nparams['AB']
   
  # this will likely only happen if two alignments are identical
  if result <= 0:
    sys.stderr.write('dlnL is less than or equal to 0! Converting to 1E-6.\n')
    result = 1e-6
  
  return (result, df)


def cleanup():
  """Check for and remove files that interfere with concaterpillar."""
  
  probfiles = ['infile', 'intree', 'outfile', 'outtree']
  
  probfiles.extend(glob.glob('*_phyml_*'))
  
  dlist = os.listdir('.')
  
  for i in probfiles:
    if i in dlist:
      sys.stderr.write('You have a file named "%s" in the working directory.\n'
                       % i)
      while True:
        answer = raw_input('Remove this file (y/n)? ')
        if answer == 'Y' or answer == 'y':
          os.remove(i)
          break
        elif answer == 'N' or answer == 'n':
          sys.stderr.write('Please move or remove any files named "infile", ' +
                           '"intree", "outfile", "outtree",\nor "*_phyml_*".\n')
          sys.exit(1)
        else:
          sys.stderr.write('Invalid response.\n')
          
          
def getalllikes(nsets):
  """Return a matrix of likelihoods from TREE-PUZZLE output files.

  The TREE-PUZZLE output files should have been produced by evaluating the
  likelihood for a data set (set A) under 2 trees: the ML tree from set A, and 
  the ML tree from a data set produced by concatenating set A with a second set
  (set B).  Files need to be named set*$.puz, where * and $ are each 3-digit 
  numbers, * representing the number assigned to set A, and $ the number
  assigned to set B.  Sets must be numbered consecutively, starting with 000.

  Parameters:
    nsets -- number of alignments for which likelihoods were reestimated

  Return:
    a matrix of pairs of likelihoods [a, b], in which a is the likelihood of 
    data set A, given the ML (estimated) tree for A, and b is the likelihood
    of data set A, given the ML (estimated) tree for data set AB.  This is a
    square matrix, in which the diagonal is missing (i.e., matrix indices
    almost correspond to data set numbers, with the row representing the first
    number in the puzzle file (* above) and the column representing the second
    ($ above), but column index is off by one (too low), where column > row).
    
  """
  likes = []
  
  for i in range(nsets):
    likes.append([])
    for j in range(nsets):
      if i != j:
        likefile = open('set%03d%03d.lnl' % (i, j))
        likes[-1].append([float(line) for line in likefile])
        likefile.close()

  return likes

def getalpha(ofname):
  """Extract the estimated alpha value from a PHYML output file.
  
  Parameters:
    ofname -- name of the PHYML output file in which to find the alpha value
    
  Return:
    Extracted alpha value (a float).
  
  """
  
  try:
    outfile = open(ofname, 'r')
  except IOError:
    raise ConcaterpillarError('Couldn\'t find PHYML file %s!' % ofname)

  alpha = None
      
  for line in outfile:
    if bool(re.search('Gamma', line)):
      wordlist = string.split(line)
      alpha = float(wordlist[-1])
      break
    
  outfile.close()
  
  if alpha == None:
    raise ConcaterpillarError('Could not find alpha in file %s!' % ofname)
  
  return alpha

def getinput():
  """Display a menu and determine at which stage Concaterpillar should begin.
  
  Return:
    Dictionary summarising values selected from the menu.
  
  """

  exstat = os.system('clear')
  if exstat != 0:
    os.system('cls')
                              
  print '\n\nWelcome to Concaterpillar!'
  print '\
                     o  o\
\n __  __  __  __  __  _\\/\
\n/  \\/  \\/  \\/  \\/  \\/ .\\\
\n\\__/\\__/\\__/\\__/\\__/\\_~/\
\n     ""  ""  ""  ""\n\n' 

  options = ('a', 'r', 't', 'i', 'b', 'c', 'p')
  menuitems = {'a': 'Build trees with PHYML', \
               'r': 'Reanalyze trees with PUZZLE', \
               't': 'Start NEW topological congruence test', \
               'i': 'Resume interrupted topological congruence test', \
               'b': 'Branch-length congruence test', \
               'c': 'Number of CPUs to use',\
               'p': 'Alpha level (p-value cutoff) for tests'}
  
  itemstates = {'a': True, 'r': True, 't': True, 'i': False, 'b': True, \
                'c': 1, 'p': 0.05}
  
  statemap = {True: 'Y', False: 'N'}

  while True:
    print 'OPTIONS:\n'
    for opt in options:
      if isinstance(itemstates[opt], types.BooleanType):
        print '%3s%60s%6s' % (opt.upper(), menuitems[opt], \
                              statemap[itemstates[opt]])
      else:
        print '%3s%60s%6s' % (opt.upper(), menuitems[opt], str(itemstates[opt]))
    
    while True:
      userin = raw_input('Item to toggle, Y to continue, Q to quit: ').lower()
    
      if userin == 'q':
        sys.exit(0)
      
      if userin == 'y':
        optdict = {}
        for key in itemstates:
          if itemstates[key] is False:
            continue
          optdict['-%s' % key] = itemstates[key]
            
        return optdict
    
      if userin == 'a':
        if itemstates['a']:
          itemstates['a'] = False
          break
        
        else:
          if itemstates['t'] and not itemstates['r']:
            print 'Either turn off option T, or turn on option R first!'
          elif itemstates['b'] and not itemstates['r']:
            print 'Either turn off option B, or turn on option T first!'
          else:
            itemstates['a'] = True
            break
            
      elif userin == 'r':
        if itemstates['r']:
          if itemstates['a'] and itemstates['t']:
            print 'Turn off either option A or option T first!'
          else:
            itemstates['r'] = False
            break
        
        else:
          if itemstates['b'] and not itemstates['t']:
            print 'Either turn off option B, or turn on option T first!'
          else:
            itemstates['r'] = True
            break
          
      elif userin == 't':
        if itemstates['t']:
          if itemstates['b'] and (itemstates['a'] or itemstates['r']):
            print 'Either turn off option B or options A and R first!'
          else:
            itemstates['t'] = False
            break
            
        else:
          if itemstates['a'] and not itemstates['r']:
            print 'Either turn on both A and R, just R, or neither first!'
          elif itemstates['i']:
            print 'Turn off option I first!'
          else:
            itemstates['t'] = True
            break
          
      elif userin == 'i':
        if itemstates['i']:
          itemstates['i'] = False
          break
        else:
          if itemstates['a'] or itemstates['r'] or itemstates['t']:
            print 'Turn off options A, R, and T first!'
          else:
            itemstates['i'] = True
          
      elif userin == 'b':
        if itemstates['b']:
          itemstates['b'] = False
          break
          
        else:
          if not itemstates['t'] and (itemstates['a'] or itemstates['r']):
            print 'Either turn on option T, or turn off A and R first!'
          else:
            itemstates['b'] = True
            break
          
      elif userin == 'c':
        nprocs = raw_input('How many CPUs do you want to use? ')
        try:
          nprocs = int(nprocs)
          if nprocs < 1:
            raise ValueError
          itemstates['c'] = nprocs
          break
        
        except ValueError:
          print 'Must use at least 1 processor!'
        
      elif userin == 'p':
        cutoff = raw_input('What cutoff do you wish to use? ')
        try:
          cutoff = float(cutoff)
          if cutoff < 0 or cutoff > 1:
            raise ValueError
          itemstates['p'] = cutoff
          break
        
        except ValueError:
          print 'Cutoff must be a value between 0 and 1!'
          
      
      else:
        print 'Invalid option!'
    
def getlikes(ofname):
  """Extract and return a list of log-likelihoods from a TREE-PUZZLE outfile.

  Parameters:
    ofname -- name of the PUZZLE output file.
    
  Return:
    Extracted log-likelihoods.

  """

  try:
    outfile = open(ofname, 'r')
  except IOError:
    raise ConcaterpillarError('Couldn\'t find PUZZLE file %s!' % (ofname))

  likes = []
  
  lnlRE = re.compile(r'log L: (-\d+.\d+)')

  for line in outfile:
    result = lnlRE.search(line)
    if result:
      likes.append(float(result.group(1)))

  outfile.close()

  return likes

def getmin(matrix):
  """Return the coordinates of the minimum value in a 2-D matrix.
  
  Parameters:
    matrix -- an upper-triangular matrix of likelihood ratios.
    
  Return:
    Tuple containing (row, col) of the minimum value, (-1, -1) if the matrix
    is empty or has only 1 dimension.
  
  """
  try:
    row, col = 0, 0
    min = matrix[row][col]
  except IndexError:
    return (-1,-1)

  for i in range(len(matrix)):
    for j in range(len(matrix[i])):
      if matrix[i][j] < matrix[row][col]:
        row = i
        col = j

  return row, col

def getoptdict(argv):
  """Return a dictionary of command-line options, given command line args.
  
  Terminate the program with exit code 1 and a usage message if invalid options
  are specified.

  Parameters:
    argv -- sys.argv[1:] (the script's argument vector, minus the
          name of the script itself)
          
  Return:
    a dictionary containing the values for options with short option flags as
    keys
  
  """
  
  goodopts = 'hartibc:p:'
  goodopttup = ('help', 'phyml', 'puzzle', 'topology', 'interrupt', 'blength',
                 'cpus=', 'pval=')
  
  try:
    opts, args = getopt.getopt(argv, goodopts, goodopttup)
  except getopt.GetoptError:
    usage(1)
    
  if len(args) != 0:
    usage(1)
    
  equivalents = []
  
  optdict = dict(opts)
  
  equivalents = [('-h', '--help'), ('-a', '--phyml'), ('-r', '--puzzle'), 
                 ('-t', '--topology'), ('-i', '--interrupt'), 
                 ('-b', '--blength'), ('-c', '--cpus'), ('-p', '--pval')]
  
  goodoptdict = {}
  
  for opt in equivalents:
    if opt[0] in optdict and opt[1] in optdict:
      usage(1)
    if opt[1] in optdict:
      goodoptdict[opt[0]] = optdict[opt[1]]
    elif opt[0] in optdict:
      goodoptdict[opt[0]] = optdict[opt[0]]
      
  return goodoptdict

def getseqlist():
  """Produce a list of alignment files in the working directory.

  All files with the extension .seq other than those that start with 'seq' or 
  'boot', and not including 'global.seq' are included in the list.
  
  Return:
    List of alignment file names.

  """
  
  seqlist = []
    
  for f in glob.glob('*.seq'):
    if f[:3] == 'set' or f == 'global.seq' or f[:4] == 'boot':
      continue
    seqlist.append(f)

  seqlist.sort()
  
  return seqlist

def mlboot(args):
  """Determine the topology test statistic by ML using a bootstrapped dataset.
  
  PHYML is run to get trees, and TREE-PUZZLE to re-estimate log-likelihoods.

  Parameters:
    args -- a dictionary containing the following items:
      'lock': a threading.Lock object used to prevent deadlocking
      'startcount': a Counter used to determine iteration number
      'align': alignment from which sites are sampled
      'ncharA': number of sites to sample for the first alignment
      'ncharB': number of sites to sample for the second alignment
      'endcount': a Counter used to determine how many bootstrap replicates 
                  have finished
      'niter': indicates total number of replicates to perform
      'teststats': list to which the test statistic is appended

  """
  
  phymlpath = globals()['phymlpath']
  puzzlepath = globals()['puzzlepath']
  
  args['lock'].acquire()
  iterno = int(args['startcount'])
  args['startcount'].increment()
  args['lock'].release()
  
  # names of bootstrap files
  bootseqs = ['boot%03d%s.seq' % (iterno, c) for c in ('a', 'b', 'ab')]
  
  writephy(bootstrap(args['align'], args['ncharA']), bootseqs[0])
  writephy(bootstrap(args['align'], args['ncharB']), bootseqs[1])  
  writephy(catseq(getaligns((bootseqs[0], bootseqs[1]))), bootseqs[2])
     
  for seq in bootseqs:
    runphyml(phymlpath, alignname=seq, model='JTT', alpha=args['alpha'])
    os.rename('%s_phyml_tree.txt' % seq, seq[:-4] + '.tre')
    os.remove('%s_phyml_stat.txt' % seq)
    os.remove('%s_phyml_lk.txt' % seq)
      
  # likelihood re-evaluation
  puzzlecmds = ['m', 'm', 'w', 'a', str(args['alpha']), 'c', '4', 'y']

  likes = []    
  trees = []
    
  for seq in bootseqs:
    treefile = open(seq[:-4] + '.tre')
    trees.append(string.strip(treefile.readline()))
    treefile.close()

  for i in (0, 1):
    intree = '%s.intree' % (bootseqs[i][:-4])
    treefile = open(intree, 'w')
    treefile.write('%s\n%s\n' % (trees[i], trees[2]))
    treefile.close()
    runbin('%s %s %s' % (puzzlepath, bootseqs[i], intree), puzzlecmds)
    likes.extend(getlikes(intree + '.puzzle')) 

  for f in glob.glob('boot%03d*' % iterno):
    os.remove(f)

  args['lock'].acquire()
  
  # should be lnL(A)_{T_A} + lnL(B)_{T_B} - (lnL(A)_{T_AB} + lnL(B)_{T_AB}))
  # because lnL(A)_{T_AB} might be greater than lnL(A)_{T_A}, max taken instead
  args['teststats'].append(max(likes[:2]) - likes[1] + \
                           max(likes[2:]) - likes[3])

  # print pretty progress counter
  countval = int(args['endcount'])
  args['endcount'].increment()
  if countval % 10 == 0:
    sys.stdout.write('[.')
  elif countval % 10 == 9:
    sys.stdout.write('.] %d/%d\n' % (countval + 1, args['niter']))
  else:
    sys.stdout.write('.')    
  sys.stdout.flush()
  
  args['lock'].release()
  
def mlpart(args):
  """Determine topology test statistic by ML with randomly partitioned data set.
  
  This function isn't actually used anymore, but was implemented as an
  alternative to mlboot (above).  It didn't work as well.  However, it's an 
  interesting enough alternative that I'm leaving the code here.

  Parameters:
    args -- a dictionary containing the following items:
      'lock': a threading.Lock object used to prevent deadlocking
      'startcount': a Counter used to determine iteration number
      'align': alignment from which sites are sampled
      'tree': tree produced from concatenated alignment
      'ncharA': number of sites to sample for the first alignment
      'ncharB': number of sites to sample for the second alignment
      'endcount': a Counter used to determine how many bootstrap replicates 
                  have finished
      'niter': indicates total number of replicates to perform
      'teststats': list to which the test statistic is appended

  """
  
  phymlpath = globals()['phymlpath']
  puzzlepath = globals()['puzzlepath']
  
  args['lock'].acquire()
  iterno = int(args['startcount'])
  args['startcount'].increment()
  args['lock'].release()
    
  # repartition concatenated alignment
  alignA, alignB = repartition(args['align'], (args['ncharA'], args['ncharB']))
  
  # use phyml to analyze both sets (estimate tree and alpha)
  writephy(alignA, 'boot%03dab.seq' % iterno)
  writephy(alignB, 'boot%03dab.seq' % iterno, append=True)
  runphyml(phymlpath, alignname='boot%03dab.seq' % iterno, nsets = 2, model='JTT')
  
  treefile = open('boot%03dab.seq_phyml_tree.txt' % iterno) 
  trees = [line.strip() for line in treefile]
  treefile.close()
  
  alphas = []
  regex = re.compile(r'#\d+')

  alphafile = open('boot%03dab.seq_phyml_stat.txt' % iterno)
  for line in alphafile:
    if regex.match(line.strip()): 
      alphas.append(float(line.split()[-2])) # next-to-last token on line
      
  alphafile.close()
  
  # use puzzle to analyze each set, with its own tree and the concatenated tree
  likes = []
  
  # placeholder will be replaced by partition-specific alpha
  puzzlecmds = ['m', 'm', 'w', 'a', 'placeholder', 'c', '4', 'y']

  alignfiles = ('boot%03da.seq' % iterno, 'boot%03db.seq' % iterno)
  
  writephy(alignA, alignfiles[0])
  writephy(alignB, alignfiles[1])
  
  for i in (0, 1):
    puzzlecmds[4] = str(alphas[i])
    
    intree = '%s.intree' % alignfiles[i]
    treefile = open(intree, 'w')
    treefile.write('%s\n%s\n' % (trees[i], args['tree']))
    treefile.close()
    
    runbin('%s %s %s' % (puzzlepath, alignfiles[i], intree), puzzlecmds)
    likes.extend(getlikes(intree + '.puzzle'))
        
  for f in glob.glob('boot%03d*' % iterno):
    os.remove(f)
  
  # calculate and return LR stat
  args['lock'].acquire()
    
  # should be lnL(A)_{T_A} + lnL(B)_{T_B} - (lnL(A)_{T_AB} + lnL(B)_{T_AB}))
  # because lnL(A)_{T_AB} might be greater than lnL(A)_{T_A}, max taken instead
  args['teststats'].append(max(likes[:2]) - likes[1] + \
                           max(likes[2:]) - likes[3])
  
  countval = int(args['endcount'])
  args['endcount'].increment()
  if countval % 10 == 0:
    sys.stdout.write('[.')
  elif countval % 10 == 9:
    sys.stdout.write('.] %d/%d\n' % (countval + 1, args['niter']))
  else:
    sys.stdout.write('.')    
  sys.stdout.flush()
  
  args['lock'].release()  

def multipartition(setA, setB, niter, nprocs):
  """Repartition data set, evaluating LRT statistic for each iteration.

  Used to get a null distribution for the topological congruence test.
  Alternative to bootstrapping, where combined data set is randomly
  repartitioned into two "genes"

  Parameters:
    setA -- number of first set
    setB -- number of second set
    niter -- number of times to repartition data set and get pseudo test stat
    nprocs -- number of processes to thread off... should be number of 
              processors to use on a shared memory machine
              
  Return:
    List of pseudo test statistics
    
  """
  
  teststats = []
  
  alignA = getseqs('set%03d.seq' % setA)
  alignB = getseqs('set%03d.seq' % setB)
  
  # want same number of taxa in bootstraps as in real dataset
  alignA, alignB = getsequnion(alignA, alignB)
  
  ncharA = alignA.length
  ncharB = alignB.length
  
  alignAB = catseq((alignA, alignB))

  # don't need these anymore
  del alignA
  del alignB
  
  treefile = open('set%03d%03d.tre' % (setA, setB))
  treeAB = treefile.readline().strip()
  treeAB = root(cleantree(treeAB))
  treeAB = prunetree(treeAB, alignAB.taxlist)
  treefile.close()
  
  treeAB = root(cleantree(treeAB))
  
  q = Queue.Queue()
  lock = threading.Lock()
  endcount = Counter() # determines how many iterations finished
  startcount = Counter() # determines how many iterations started
  threadcounter = Counter() 
  
  for i in range(niter):
    argsdict = {}
    argsdict['lock'] = lock
    argsdict['ncharA'] = ncharA
    argsdict['ncharB'] = ncharB
    argsdict['endcount'] = endcount
    argsdict['startcount'] = startcount
    argsdict['niter'] = niter
    argsdict['teststats'] = teststats

    argsdict['align'] = alignAB
    argsdict['tree'] = treeAB
    
    q.put(argsdict)
    
  for i in range(nprocs):
    t = PhyloThread(q, lock, target = mlpart, name = 'thread %d' % i,
                      counter = threadcounter)
    t.start()
          
    # wait until new thread really starts before starting next thread
    lock.acquire()
    lock.release()
    
  while int(threadcounter) > 0:
    time.sleep(10)
    
  return teststats

def multinonparboot(setA, setB, niter, nprocs):
  """Perform several ML-based non-parametric bootstrap replicates  
  
  Parameters:
    setA -- number of first set
    setB -- number of second set
    niter -- number of times to repartition data set and get pseudo test stat
    nprocs -- number of processes to thread off... should be number of 
              processors to use on a shared memory machine
              
  Return:
    List of pseudo test statistics

"""
      
  teststats = []
  
  alignA = getseqs('set%03d.seq' % setA)
  alignB = getseqs('set%03d.seq' % setB)
  
  # want same # of taxa in bootstraps as in real dataset
  alignA, alignB = getsequnion(alignA, alignB)
  
  # comment these 2 lines out to resample from concatenated set
  alphaA = readalpha('set%03d.alpha' % setA)
  alphaB = readalpha('set%03d.alpha' % setB)
  
  ncharA = alignA.length
  ncharB = alignB.length
  
  q = Queue.Queue()
  lock = threading.Lock()
  endcount = Counter() # determines how many iterations finished
  startcount = Counter() # determines how many iterations started
  threadcounter = Counter() 
  
  for i in range(niter):
    argsdict = {}
    argsdict['lock'] = lock
    argsdict['ncharA'] = ncharA
    argsdict['ncharB'] = ncharB
    argsdict['endcount'] = endcount
    argsdict['startcount'] = startcount
    argsdict['niter'] = niter
    argsdict['teststats'] = teststats

    # comment out this block to resample from concatenated set
    if i < niter/2: 
      argsdict['alpha'] = alphaA
      argsdict['align'] = alignA
    else:
      argsdict['alpha'] = alphaB
      argsdict['align'] = alignB

    # uncomment these 2 lines to resample from concatenated set
    #argsdict['align'] = catseq((alignA, alignB))
    #argsdict['alpha'] = readalpha('set%03d%03d.alpha' % (setA, setB))
      
    q.put(argsdict)
    
  for i in range(nprocs):
    t = PhyloThread(q, lock, target = mlboot, name = 'thread %d' % i,
                      counter = threadcounter)
    t.start()
          
    # wait until new thread really starts before starting next thread
    lock.acquire()
    lock.release()
    
  while int(threadcounter) > 0:
    time.sleep(10)
    
  return teststats

def preparesets(seqfiles):
  """Concatenate all pairs of files in seqfiles, and return the new file names.

  Copy all files in seqfiles to consecutively numbered files with names of the
  form set*.seq, where * is a 3 digit number, and prepare pairwise concetenated
  sequence files with names of the form set*$.seq, where * is the 3 digit number
  assigned to the first file and $ is the 3 digit number assigned to the second.
  Return a list of the new files.
  
  Parameters:
    seqfiles -- a list of names of files to be copied/concatenated
    
  Return:
    List of file names for new concatenated data sets.
  
  """
  sets = []
  
  for i in range(len(seqfiles)):
    print 'Reading %s...' % seqfiles[i]
    writephy(getseqs(seqfiles[i]), 'set%03d.seq' % i)
    sets.append('set%03d.seq' % i)
          
  nsets = len(seqfiles)
    
  for i in range(nsets):
    for j in range(i + 1, nsets):
      writephy(catseq(getaligns((sets[i], sets[j]))), 'set%03d%03d.seq' 
               % (i, j))
      sets.append('set%03d%03d.seq' % (i, j))
            
  return sets

def readalpha(alphafile):
  """Read the value from an "alpha" file (contains a single float).

  Parameters:
    alphafile -- name of the file from which the alpha value is to be read
    
  Return:
    Alpha value read from alphafile (as a float).
  
  """
  try:
    af = open(alphafile)
    alpha = float(af.readline())
    af.close()
    
  except IOError:
    raise ConcaterpillarError('Alpha file %s could not be found!' % alphafile)
  except ValueError:
    raise ConcaterpillarError('File %s does not appear to be an alphafile!' \
                              % alphafile)
  
  return alpha

def recoversets():
  """Produce a list containing the names of concatenated alignments.

  The list returned is that recovered from the backup file catsets.pickle.  It
  contains lists, each of which contains the original names of alignment files
  that have been concatenated (clustered) as a set.
  
  Return:
    List of lists containing names of sets that have already been concatenated.
  
  """
  
  try:
    backup = open('catsets.pickle')
  except IOError:
    sys.stderr.write('Could not resume congruence test.  Sorry.\n')
    sys.exit(1)
      
  catsets = cPickle.load(backup)
  backup.close()
  
  setcount = 0
  for set in glob.glob('set*.seq'):
    if len(set) == 10:
      setcount += 1
      
  if setcount != len(catsets):
    sys.stderr.write('Could not resume congruence test.  Sorry.\n')
    sys.exit(1)
    
  return catsets

def removesets(setid):
  """Remove all concatenated set files with setid as one of the set ID #s.

  The alignment files as well as any files produced from the alignment files
  (tree files, alpha files, etc.) are deleted.

  Parameters:
    setid -- identification number (int) for concatenated sets to be deleted.

  """
  
  setlist = []
  setRE1 = re.compile(r'set\d\d\d%03d' % setid)
  setRE2 = re.compile(r'set%03d\d\d\d' % setid)

  for f in os.listdir('.'):
    if setRE1.match(f) or setRE2.match(f):
      setlist.append(f)
      
  setlist.sort()
  
  removefiles(setlist)

def renamesets(setid):
  """Rename sets with IDs greater than setid.

  Remove sets that contain the set with id setid (sets that SHOULD be 
  useless at this point), and rename sets with IDs greater that setid (i.e.,
  if setid = 3, set003 sets should be removed, and set004 becomes set003, set005
  becomes set004, etc.)  In other words, shift set numbers down by 1.

  Parameters:
    setid -- ID number of the set above which sets are renamed.
  """

  namesdict = {}
  
  # concatRE will be something like "set000123"
  concatRE = re.compile(r'set(\d\d\d)(\d\d\d)')
  simpleRE = re.compile(r'set(\d\d\d)')
  
  for f in glob.glob('set*'):#allsetfiles:
    result = concatRE.match(f)
    
    if result: # set is concatenated pair (two ID numbers)
      num1, num2 = [int(i) for i in result.groups()]
    
      # don't need to rename sets that fall below setid in both cases
      if (num1 < setid) and (num2 < setid):
        continue
    
      if num1 > setid:
        num1 -= 1
      if num2 > setid:
        num2 -= 1
        
      suffix = f.split('.')[-1]
    
      # add an entry to namesdict to indicate how to rename the set
      namesdict[f] = 'set%03d%03d.%s' % (num1, num2, suffix)

    # set has a single ID number
    else: 
      result = simpleRE.match(f)
      try:
        num = int(result.group(1))
      except (ValueError, IndexError):
        raise ConcaterpillarError('Filename %s is not a legitamate set name!' \
                                  % f)
      except AttributeError:
        continue
      
      if num < setid:
        continue
      suffix = f.split('.')[-1]
      namesdict[f] = 'set%03d.%s' % (num - 1, suffix)
      
  renamefiles(namesdict)

def runpuzzlebrlen(argsdict):
  """Run TREE-PUZZLE to get likelihoods required for branch-length test.
  
  Parameters:
    argsdict -- dictionary containing the following items:
      'tree': Newick-format tree string (with branch lengths, whatever)
      'seqfile': name of seqfile
      'gammacats': integer, number of categories for discrete Gamma distribution
                   (should probably just be 4)
  
  """
  
  tree = argsdict['tree']
  seqfile = argsdict['seqfile']
  gammacats = str(argsdict['gammacats'])
  prefix = seqfile[:-4]
  
  if gammacats == 4 and prefix + '.lk' in os.listdir('.'):
    print 'Already done TREE-PUZZLE analysis of %s.' % seqfile
    return
  puzzlecmds = ['', 'e', 'm', 'm', 'w', 'c', gammacats, 'y']
  taxlst = getseqs(seqfile).taxlist
  
  # number of branch lengths + 1
  np = len(taxlst) * 2 - 2 
  paramfile = open('%s.prm' % seqfile[:-4], 'w')
  paramfile.write('%d\n' % np) 
  paramfile.close()
      
  treefile = open(prefix + '.tre', 'w')
  treefile.write('%s\n' % unroot(prunetree(tree, taxlst)))
  treefile.close()
      
  runbin('%s %s %s.tre -prefix=%s' % (puzzlepath, seqfile, prefix, prefix), puzzlecmds)
  print 'Finished TREE-PUZZLE analysis of %s.' % seqfile
  
  # extract & save likelihood from puzzle outfile
  likefile = open('%s.lnl' % prefix, 'w')
  likefile.write('%f\n' % getlikes('%s.puzzle' % prefix)[0])
  likefile.close()
  
  # delete puzzle outfiles
  for ext in ('dist', 'tree', 'puzzle'):
    os.remove('%s.%s' % (prefix, ext))
    

def toptest(catsets, likes, cutoff, nprocs = 1):#, outfile, statsfile):
  """Execute the topological congruence test.

  Parameters:
    catsets -- list of lists containing sets that have already been concatenated
    likes -- matrix of pairs of likelihoods (as described in the documentation
             for buildmatrix, above)
    cutoff -- uncorrected alpha level for the test 
    
  """
  
  backup = open('catsets.pickle', 'w')
  matrix = buildmatrix(likes)
  matfile = open('matrices.tab', 'a')
  writematrix(matrix, matfile)
  outfile = open('results.ccp', 'a')
    
  nsets = 0
  for set in catsets:
    nsets += len(set)
    
  # current level is number of concatenations that have alredy happened (single-
  # gene sets minus remaining sets plus 1... note incrementation in loop below)
  level = nsets - len(catsets)
  
  while len(matrix) > 0:
    level += 1
    row, col = getmin(matrix)
    
    setno1 = row
    setno2 = col + row + 1 # adjust for triangular matrix
      
    sys.stdout.write('\nStarting bootstrap replicates:\n')
    niter = 100
    
    # Change to multipartition call, if partitioning prefered to bootstrapping
    results = multinonparboot(setno1, setno2, niter, nprocs)    
    sys.stdout.write('\n')
    results.sort()
    
    # Uncomment to record bootstrap test stats (and real test stat) to a file
    #bootfile = open('bootvals', 'a')
    #bootfile.write('%f ' % matrix[row][col])
    #for r in results:
    #  bootfile.write('%f ' % r)
    #bootfile.write('\n')


    pval = getpval(matrix[row][col], results)
    rawp = getrawpval(matrix[row][col], results)
    print 'Raw p-value: %.2f' % rawp
    print 'Weibull-smoothed p-value: %f' % pval
    
    # Note: the original correction method was to "predict" the number of levels
    # using the uncorrected alpha, then to correct based on this value.  But
    # since the correction is used WHENEVER the null is rejected with the
    # uncorrected alpha, it's simpler to just always correct based on the
    # minimum number of levels... i.e., the current level.  This works because
    # the corrected alpha is always less than or equal to the user-defined alpha
    
    alpha = correctalpha(cutoff, level)

    # ... but only tell the user we're correcting when we NEED to correct
    if pval < cutoff:
      print 'Applying alpha-level correction: new alpha = %.6f' % alpha
      outfile.write('Applying alpha-level correction: new alpha = %.6f\n' % alpha)
      sys.stdout.flush()
      outfile.flush()
      
    # if null is rejected
    if pval < alpha:
      print 'Failed to concatenate %s and %s (p = %.6f).\n' % \
              (str(catsets[setno1]), str(catsets[setno2]), pval)
      outfile.write('\nFailed to concatenate %s and %s (p = %.6f).\n\n' % \
                    (str(catsets[setno1]), str(catsets[setno2]), pval))
      outfile.flush()
      break    
    
    # otherwise...
    print 'Concatenating %s and %s (p =  %.6f).\n' % \
          (str(catsets[setno1]), str(catsets[setno2]), pval)
    outfile.write('Concatenating %s and %s (p = %.6f).\n' % \
                  (str(catsets[setno1]), str(catsets[setno2]), pval))
    outfile.flush()
        
    namesdict = {}
    removefiles(glob.glob('set%03d.*' % setno1))

    for suffix in ('seq', 'tre', 'alpha'):
      namesdict['set%03d%03d.%s' % (setno1, setno2, suffix)] = \
              'set%03d.%s' % (setno1, suffix)
    renamefiles(namesdict)
    
    # only necessary if setno2 is set with highest ID
    removefiles(glob.glob('set%03d.*' % setno2))
    removesets(setno1)
    removesets(setno2)
    renamesets(setno2)
    
    catsets[setno1].extend(catsets[setno2])
    del catsets[setno2]  

    # concatenate necessary data sets
    nsets = len(catsets)
    newsets = []
    for setid in range(nsets):
      if setid < setno1:
        setname = 'set%03d%03d.seq' % (setid, setno1)
      elif setid > setno1:
        setname = 'set%03d%03d.seq' % (setno1, setid)
      else:
        continue
      writephy(catseq(getaligns(('set%03d.seq' % setid, 
                                 'set%03d.seq' % setno1))), setname)
      newsets.append(setname)
    
    # save resume point
    backup.seek(0)
    cPickle.dump(catsets, backup)
    backup.flush()
    backup.truncate()
    
    # analyze new data sets
    if newsets:
      buildmultitrees(newsets, nprocs)
      runmultipuzzle(setnames = newsets, nprocs = nprocs)
    
      likes = getalllikes(len(catsets))
      matrix = buildmatrix(likes)
      writematrix(matrix, matfile)
      
    else:
      break

  backup.close()
  matfile.close()
  
  print 'Finished concatenating sets.  The following files contain the',
  print 'final sets:'
  outfile.write('Finished concatenating sets.  ')
  outfile.write('The following files contain the final sets:\n')

  for i in range(len(catsets)):
    print 'set%03d.seq (%d genes): %s' % (i, len(catsets[i]), str(catsets[i]))
    outfile.write('set%03d.seq (%d genes): %s\n' \
                  % (i, len(catsets[i]), str(catsets[i])))

  outfile.close()

def usage(exitcode):
  """Print a usage message and exit with the given exit code."""
  
  from concaterpillar import __doc__
  
  sys.stderr.write(__doc__)
  sys.exit(exitcode)

def writematrix(matrix, matfile):
  """Write likelihood ratio matrix to a file in PHYLIP distance format.
  
  Parameters:
    matrix -- log-likelihood ratio matrix
    matfile -- file to record distance matrix to
  
  """
  
  nsets = len(matrix)
  
  matfile.write('  %d\n' % nsets)
  for i in range(nsets + 1):
    matfile.write('set%03d   ' % i)
    for j in range(nsets + 1):
      if i == j:
        val = 0.0
      elif i < j:
        val = matrix[i][j - i - 1]
      else:
        val = matrix[j][i - j - 1]
        
      matfile.write('%10.2f' % val)
        
    matfile.write('\n')
    matfile.flush()

class ConcaterpillarError(Exception):  
  """
  Exception class designed for concaterpillar-specific errors.
  
  Extends Exception.
  
  """
  
  def __init__(self, value):
    self.value = value
    
  def __str__(self):
    return repr(self.value)

    
class Counter:
  """
  Implementation of an integer object that can be passed as a reference.
    
  """
  
  def __init__(self, initval = 0):
    self.count = initval
    
  def increment(self):
    self.count += 1
  
  def decrement(self):
    self.count -= 1
    
  def __int__(self):
    return self.count
    
  def __repr__(self):
    return str(self.count)
  
  def __str__(self):
    return repr(self)

class PhyloThread(threading.Thread):
  """
  Thread that runs an appropriate target phylogeny programme.
  
  Extends threading.Thread
  
  """
  
  def __init__(self, jobqueue, lock, target, counter, name = None):
    """Create a PhyloThread object.
    
    Parameters:
      jobqueue -- queue from which job objects are drawn by all active threads
      lock -- shared lock object to prevent deadlocking of threads
      target -- phylogeny programme to run
      counter -- Counter object to count active threads
      name -- name assigned to thread (doesn't really DO anything)
    
    """
    
    threading.Thread.__init__(self, name = name)
    self.name = name
    self.jobqueue = jobqueue
    self.lock = lock
    self.target = target
    self.counter = counter
    
    self.lock.acquire()
    self.counter.increment()
    self.lock.release()
    
  def run(self):
    while True:
      self.lock.acquire()
      
      if self.jobqueue.empty():
        self.lock.release()
        break
      
      arg = self.jobqueue.get()
      self.lock.release()
      
      self.target(arg)
 
    self.lock.acquire() 
    self.counter.decrement()
    self.lock.release()      

def runmultipuzzle(nsets = None, setnames = None, nprocs = 1):
  """Launch multiple threads to run TREE-PUZZLE for toptest.
  
  This function should be called with either nsets or setnames, but not both.
  If setnames is specified, the numbers of the sets in the setnames list will
  be extracted and passed to runpuzzle.  If nsets is specified, the appropriate
  set numbers will be generated.
  
  Parameters:
    nsets -- integer specifying the number of (single) sets to be reanalysed.
    setnames -- names of sets to be reanalysed.
    nprocs -- number of processes to thread off... should be number of 
              processors to use on a shared memory machine
  
  """  
  
  print 'Starting TREE-PUZZLE reanalysis of trees.'
  
  q = Queue.Queue()
  lock = threading.Lock()
  threadcounter = Counter()
  
  try:
    if setnames:
      for name in setnames:
        i = int(name[3:6])
        j = int(name[6:9])
        q.put((i, j))
        q.put((j, i))
        
    elif nsets:
      for i in range(nsets):
        for j in range(nsets):
          if i != j:
            q.put((i, j))
    
    else:
      sys.stderr.write('runmultipuzzle was not called properly.\n%s' % \
                       runmultipuzzle.__doc__)
      sys.exit(1)
            
  except IOError: 
    sys.stderr.write('Invalid starting conditions for tree reanalysis.\n')
    usage(1)
          
  for i in range(nprocs):
    t = PhyloThread(q, lock, runpuzzle, name = 'thread %d' % i, 
                    counter = threadcounter)
    t.start()
    
    # wait until newly-started thread really starts before starting next thread
    lock.acquire()
    lock.release()
    
  while int(threadcounter) > 0:
    time.sleep(10) 
    
  print 'Finished reanalysis.'

def runpuzzle(seqids):
  """Run puzzle on a set with trees from single and concatenated sets.
  
  Taxa may be pruned from trees if the two data sets don't share the same taxa.
  The tree inferred from the data set with seqids[0] is only reanalysed if taxa
  were pruned, or if the unpruned tree wasn't already reanalysed.
  
  Parameters:
    seqids -- tuple containing two set ID numbers: the first is the set for
              which likelihoods will be determined, the second is the set with
              which set seqids[0] was concatenated to get the second tree to be
              evaluated
              
  Return:
    Log-likelihoods for trees evaluated with the alignment in question.
    
  """

  puzzlepath = globals()['puzzlepath']
  
  seqid1, seqid2 = seqids
    
  tf1 = open('set%03d.tre' % seqid1)
  tree1 = string.strip(tf1.readline())
  tf1.close()
   
  tree1 = root(cleantree(tree1))
    
  if seqid1 < seqid2:
    tf2 = open('set%03d%03d.tre' % (seqid1, seqid2))
  else:
    tf2 = open('set%03d%03d.tre' % (seqid2, seqid1))
    
  tree2 = string.strip(tf2.readline())
  tf2.close()
  tree2 = root(cleantree(tree2))
              
  align1 = getseqs('set%03d.seq' % seqid1)
  align2 = getseqs('set%03d.seq' % seqid2)
  
  taxa1 = align1.taxlist[:]
  taxa1.sort()
  taxa2 = align2.taxlist[:]
  taxa2.sort()
  
  foundlike = False

  # If the two alignments share the same taxa, and tree1 has already been 
  # reevaluated, take note.
  if taxa1 == taxa2:
    sametaxa = True
    try:
      likefile = open('set%03d.lnl' % seqid1)
      while True:
        try:
          lnL1 = float(likefile.readline())
          likefile.close()
          foundlike = True
          break
        except ValueError:
          pass
      
    except IOError:
      pass
    
  else:
    sametaxa = False
    align1, align2 = getsequnion(align1, align2)
  
  alphafile = open('set%03d.alpha' % seqid1)
  alpha = alphafile.readline().strip()
  alphafile.close()

  intreename = 'set%03d%03d.puztre' % (seqid1, seqid2)
  treefile = open(intreename, 'w')

  if foundlike:
    treefile.write('%s\n' % tree2)
    
  # No likelihood yet calculated for this tree with these taxa
  else:
    treefile.write('%s\n%s\n' % (prunetree(tree1, align1.taxlist), 
                                 prunetree(tree2, align1.taxlist)))
  treefile.close()
  
  infilename = 'set%03d%03d.puzin' % (seqid1, seqid2)
  writephy(align1, infilename)
  
  puzzlecmds = ['', 'e', 'm', 'm', 'w', 'c', '4', 'a', alpha, 'y']
  runbin('%s %s %s' % (puzzlepath, infilename, intreename), puzzlecmds) 
  
  likes = getlikes('%s.puzzle' % intreename)
  
  if foundlike:
    likes[:0] = [lnL1]
  elif sametaxa:
    likefile = open('set%03d.lnl' % seqid1, 'w')
    likefile.write('%f\n' % likes[0])
    likefile.close()
    
  likefile = open('set%03d%03d.lnl' % (seqid1, seqid2), 'w')
  likefile.write('%f\n%f\n' % tuple(likes))
  
  return likes
  
try:
  from concaterpillar import phymlpath
except ImportError:
  phymlpath = getpath('phyml')
  
try:
  from concaterpillar import puzzlepath
except ImportError:
  puzzlepath = getpath('puzzle')
