#!/usr/bin/python
"""
concaterpillar.py

Assessment of congruence for protein alignment concatenation.

Usage: python concaterpillar.py [options]

  Options:
    -h, --help                print this message
    -a, --phyml               begin building trees with PHYML
    -r, --puzzle              begin reanalysis of trees with TREE-PUZZLE
    -t, --topology            begin topological congruence test
    -i, --interrupt           continue interrupted topological congruence test
    -b, --blength             begin branch length compatibility assessment
    -c ncpu, --cpus=ncpu      use specified number of processors (ncpu must be 
                              integer greater than 0, default = 1)
    -p cutoff, --pval=cutoff  use specified p-value as a cutoff for tests 
                              (must be a number between 0 and 1, default = 0.05)
    
  Notes:
    At least one option of -a, -r, -t, -i, and -b must be specified.  More
    than one may be used.  Note also that if you run with one of these switches
    at a time, -a must be run before -r, and -r before -t.  -i should only be
    used if trees have already been build with PHYML (-a) and evaluated with
    TREE-PUZZLE (-r), and the topology test (-t) has begun, but has been
    interrupted prior to completion for some reason.  In theory, -b can be used
    independently of all other options, in which case branch length 
    compatibility of all alignments will be evaluated.  Otherwise, if congruence
    test has been completed, the branch length compatibility test will be
    performed only on the alignments in the largest cluster.

"""

# TODO: rather than choosing best likelihood ratio, choose best NON-CONFLICTING
# LR. E.g., if A and B fail to pass the test, concatenated anyway; Then, if ABC
# have the best LR at the next step, ignore, but test CD instead if it's the
# next-best.  Keep track of all concatenations that pass the test, and make all
# concatenations at the end.

import cPickle, getopt, glob, shutil, sys
from ccptools import *


# Uncomment (and modify, if necessary) the following to use a specific path for 
# PHYML and TREE-PUZZLE respectively
# phymlpath = '/usr/local/bin/phyml'
# puzzlepath = '/usr/local/bin/puzzle'

if __name__ == '__main__':
    
  cleanup()
  
  # get and process command line options
  optdict = getoptdict(sys.argv[1:])

  if '-h' in optdict:
    usage(0)
    
  coreopts = ('-a', '-r', '-t', '-i', '-b')
  noopt = True
  
  for opt in coreopts:
    if opt in optdict:
      noopt = False
      break
  
  if noopt:
    # change this to a menu!
    optdict = getinput()
  
  try:
    if '-c' in optdict:
      nprocs = int(optdict['-c'])
    else:
      nprocs = 1
      
    if nprocs < 1:
      raise ValueError
      
    if '-p' in optdict:
      cutoff = float(optdict['-p'])
    else:
      cutoff = 0.05
    
    if cutoff > 1 or cutoff < 0:
      usage(1)
  
  except ValueError:
    usage(1)
    
  seqlist = getseqlist()
  
  # run phyml
  if '-a' in optdict:    
    sets = preparesets(seqlist)
    buildmultitrees(sets, nprocs)

  # run puzzle
  if '-r' in optdict:
    naligns = len(seqlist)
    runmultipuzzle(naligns, nprocs = nprocs)
    
  # resume interrupted congruence test  
  if '-i' in optdict:
    if '-a' in optdict or '-r' in optdict or '-t' in optdict:
      usage(1)
      
    catsets = recoversets() 
    compoundsets = []
    nsets = len(catsets)

    for i in range(nsets):
      for j in range(i + 1, nsets):
        compoundsets.append('set%03d%03d.seq' % (i, j))
    buildmultitrees(compoundsets, nprocs)

    runmultipuzzle(nsets, nprocs = nprocs)
    likes = getalllikes(nsets)
    
    toptest(catsets, likes, cutoff, nprocs)
  
  # start new congruence test
  if '-t' in optdict:
    if '-a' in optdict and '-r' not in optdict:
      usage(1)
    
    catsets = [[set] for set in seqlist]
    print '\nStarting compatibility analysis.'
    resfile = open('results.ccp', 'w')
    resfile.write('Concaterpillar results:\n\n')
    resfile.close()
    
    matfile = open('matrices.tab', 'w')
    matfile.close()
    
    likes = getalllikes(len(seqlist))
    try:
      toptest(catsets, likes, cutoff, nprocs)
    except IOError: 
      sys.stderr.write('Invalid starting conditions for congruence test.\n')
      raise
        
  # test branch length compatibility
  if '-b' in optdict:
    if '-t' not in optdict and '-i' not in optdict:
      if '-a' in optdict or '-r' in optdict:
        usage(1)

    allcatsetslist = brlensetup()

    if len(allcatsetslist) == 1 and 'BLtest-set000' not in os.listdir('.'):
      bltestall = True
    else:
      bltestall = False
    
    setcount = 0
    
    for catsets in allcatsetslist:
      catsets = [[set] for set in catsets]
      if bltestall:
        setdir = ''
      else:
        setdir = 'BLtest-set%03d/' % setcount
        os.chdir(setdir)
        print 'Starting branch-length congruence test on set%03d.' % setcount
      
      for i in range(len(catsets)):
        writephy(getseqs(catsets[i][0]), 'set%03d.seq' % i)
        
      fixtreefile('global.tre', roottree = True, clean = True)        
      brlentest(catsets, setdir, cutoff, nprocs)
  
      if not bltestall:
        os.chdir('..')
        setcount += 1
      
